package rpc

import (
	"sync"

	"github.com/valyala/fasthttp"

	cRPC "git.loafle.net/commons_go/rpc"
	"git.loafle.net/commons_go/rpc/protocol/json"
	cwf "git.loafle.net/commons_go/websocket_fasthttp"
	oogw "git.loafle.net/overflow/overflow_gateway_websocket"
	"git.loafle.net/overflow/overflow_gateway_websocket/rpc"
	"git.loafle.net/overflow/overflow_gateway_websocket/servlet"
)

func NewRPCGatewayServletHandler(gwRPCHandler rpc.GatewayRPCHandler) RPCGatewayServletHandler {
	gwRPCSH := newGatewayRPCServletHandler()

	sh := &RPCGatewayServletHandlers{
		gwRPCSH:      gwRPCSH,
		gwRPCHandler: gwRPCHandler,
	}

	return sh
}

type RPCGatewayServletHandlers struct {
	servlet.ServletHandlers

	gwRPCSH      GatewayRPCServletHandler
	gwRPCHandler rpc.GatewayRPCHandler
}

func (sh *RPCGatewayServletHandlers) Init(serverCTX cwf.ServerContext) error {
	if err := sh.ServletHandlers.Init(serverCTX); nil != err {
		return err
	}

	if err := sh.gwRPCHandler.Init(serverCTX); nil != err {
		return err
	}

	return nil
}

func (sh *RPCGatewayServletHandlers) Handshake(serverCTX cwf.ServerContext, ctx *fasthttp.RequestCtx) (id string, extensionsHeader *fasthttp.ResponseHeader) {
	return sh.gwRPCHandler.Handshake(serverCTX, ctx)
}

func (sh *RPCGatewayServletHandlers) SocketContext(serverCTX cwf.ServerContext) cwf.SocketContext {
	socketCTX := sh.SocketContext(serverCTX)

	return newServletContext(socketCTX)
}

// OnConnect invoked when client is connected
// If you override ths method, must call
func (sh *RPCGatewayServletHandlers) OnConnect(soc cwf.Socket) {
	sh.ServletHandlers.OnConnect(soc)

	soc.Context().SetAttribute(oogw.SocketIDKey, soc.ID())
	soc.Context().SetAttribute(cRPC.ContentTypeKey, json.Name)
}

func (sh *RPCGatewayServletHandlers) Handle(soc cwf.Socket, stopChan <-chan struct{}, doneChan chan<- error) {
	var err error
	rpcServlet := retainRPCServlet(sh.gwRPCSH)

	defer func() {
		releaseRPCServlet(rpcServlet)
		doneChan <- err
	}()

	rpcDoneChan := make(chan error, 1)

	if err = rpcServlet.Start(soc.Context(), soc, rpcDoneChan); nil != err {
		return
	}

	select {
	case err = <-rpcDoneChan:
	case <-stopChan:
		rpcServlet.Stop()
	}
}

// OnDisconnect invoked when client is disconnected
// If you override ths method, must call
func (sh *RPCGatewayServletHandlers) OnDisconnect(soc cwf.Socket) {

	sh.ServletHandlers.OnDisconnect(soc)
}

// Destroy invoked when server is stopped
// If you override ths method, must call
func (sh *RPCGatewayServletHandlers) Destroy() {

	sh.ServletHandlers.Destroy()
}

func (sh *RPCGatewayServletHandlers) Validate() {
	sh.ServletHandlers.Validate()

	sh.gwRPCHandler.Validate()
}

var rpcServletPool sync.Pool

func retainRPCServlet(sh GatewayRPCServletHandler) cRPC.Servlet {
	v := rpcServletPool.Get()
	if v == nil {
		return cRPC.NewServlet(sh)
	}
	return v.(cRPC.Servlet)
}

func releaseRPCServlet(s cRPC.Servlet) {

	rpcServletPool.Put(s)
}
